import EventEmitter from 'events';
import { writeFile } from 'fs';
import makethen from 'makethen';
import Element from './Element';
const pWriteFile = makethen(writeFile);

class Page extends EventEmitter {
  constructor(arg) {
    super();
    this._browser = arg.browser;
    this._id = arg.id;
    this._send = arg.send;
  }

  async $(selector) {
    try {
      const {
        value
      } = await this._send('WebDriver:FindElement', {
        value: selector,
        using: 'css selector'
      });
      return new Element({
        id: value,
        send: this._send
      });
    } catch (err) {
      if (err.message.startsWith('Unable to locate element')) {
        return null;
      }

      throw err;
    }
  }

  async $$(selector) {
    const values = await this._send('WebDriver:FindElements', {
      value: selector,
      using: 'css selector'
    });
    return values.map(value => new Element({
      id: value,
      send: this._send
    }));
  }

  async bringToFront() {
    return this._send('WebDriver:SwitchToWindow', {
      name: this._id,
      focus: true
    });
  }

  browser() {
    return this._browser;
  }

  async close() {
    await this._send('WebDriver:ExecuteScript', {
      script: 'window.close()'
    });
  }

  async content() {
    const {
      value
    } = await this._send('WebDriver:GetPageSource');
    return value;
  }

  async evaluate(target, ...args) {
    if (typeof target === 'function') {
      const {
        value: result
      } = await this._send('WebDriver:ExecuteAsyncScript', {
        script: `
          const args = Array.prototype.slice.call(arguments, 0, arguments.length - 1)
          const resolve = arguments[arguments.length - 1]

          Promise.resolve()
            .then(() => (${target.toString()})(...args))
            .then((value) => resolve({ error: null, value }))
            .catch((error) => resolve({ error: error instanceof Error ? error.message : error }))
        `,
        args
      });

      if (result.error !== null) {
        throw new Error(`Evaluation failed: ${result.error}`);
      }

      return result.value;
    }

    const {
      value: result
    } = await this._send('WebDriver:ExecuteAsyncScript', {
      script: `
        const resolve = arguments[0]

        Promise.resolve()
          .then(() => ${target})
          .then((value) => resolve({ error: null, value }))
          .catch((error) => resolve({ error: error instanceof Error ? error.message : error }))
      `
    });

    if (result.error !== null) {
      throw new Error(`Evaluation failed: ${result.error}`);
    }

    return result.value;
  }

  async focus(selector) {
    return this.evaluate(sel => {
      const el = document.querySelector(sel);

      if (el === null) {
        throw new Error('Unable to find element');
      }

      if (!(el instanceof HTMLElement)) {
        throw new Error('Found element is not HTMLElement and not focusable');
      }

      el.focus();
    }, selector);
  }

  async goto(url) {
    await this._send('WebDriver:Navigate', {
      url
    });
  }

  async screenshot(options = {}) {
    const result = await this._send('WebDriver:TakeScreenshot', {
      full: true,
      hash: false
    });
    const buffer = Buffer.from(result.value, 'base64');

    if (typeof options.path === 'string') {
      await pWriteFile(options.path, buffer, {
        encoding: null
      });
    }

    return buffer;
  }

  async setContent(html) {
    return this._send('WebDriver:ExecuteScript', {
      script: 'document.documentElement.innerHTML = arguments[0]',
      args: [html]
    });
  }

  async setViewport({
    width,
    height
  }) {
    const {
      widthDelta,
      heightDelta
    } = await this.evaluate(`
      ({
        widthDelta: window.outerWidth - window.innerWidth,
        heightDelta: window.outerHeight - window.innerHeight
      })
    `);
    await this._send('WebDriver:SetWindowRect', {
      width: width + widthDelta,
      height: height + heightDelta
    });
  }

  async title() {
    const result = await this._send('WebDriver:GetTitle');
    return result.value;
  }

  async url() {
    const result = await this._send('WebDriver:GetCurrentURL');
    return result.value;
  }

  async viewport() {
    return await this.evaluate(`
      ({
        width: window.innerWidth,
        height: window.innerHeight
      })
    `);
  }

}

export default Page;